import sys, getopt, os, os.path, string
from Synopsis import Type, Util, Visitor, AST

strip = []

def usage():
    print \
"""Linker Usage:
 -s <Scope>    - Select only the named scope.
                 May be used multiple times.
		 Default is to select everything.
		 Nested scopes are separated by colons.
		 Eg: -s Prague -s Berlin::Console
"""

def __parseArgs(args):
    global strip

    try:
        opts,remainder = getopt.getopt(args, "s:h")
    except getopt.error, e:
        sys.stderr.write("Error in arguments: " + e + "\n")
	sys.stderr.write("Use -Wl,-h for help")
        sys.exit(1)

    for opt in opts:
        o,a = opt

        if o == "-s":
            strip.append(a)
	elif o == "-h":
	    usage()
	    sys.exit(0)

class Stripper(Visitor.AstVisitor, Visitor.TypeVisitor):
    def __init__(self, scope):
        self.__scope = string.split(scope, "::")
    def strip(self, name):
        "strips of the scope from the name if contained"
        # flag = len(name) > len(self.__scope)
        # for i in range(0, len(self.__scope)):
        #     flag = flag and name[i] == self.__scope[i]
        # if flag:
        #     for i in range(0, len(self.__scope)):
        #         name.pop(0)
	## New version by Chalky:
	depth = len(self.__scope)
	if name[0:depth] == self.__scope:
	    del name[0:depth]
    def visitForward(self, node): return
    def visitDeclarator(self, node):
        self.strip(node.name())
    def visitModule(self, module):
        self.strip(module.name())
        for declaration in module.declarations(): declaration.accept(self)
    def visitClass(self, clas):
        self.strip(clas.name())
        for declaration in clas.declarations(): declaration.accept(self)
    def visitTypedef(self, typedef):
        for d in typedef.declarators(): d.accept(self)
    def visitEnumerator(self, enumerator):
         self.strip(enumerator.name())
    def visitEnum(self, enum):
        self.strip(enum.name())
        for e in enum.enumerators(): e.accept(self)
    def visitVariable(self, node): return
    def visitConst(self, node): return
    def visitParameter(self, node): return
    def visitFunction(self, function):
        self.strip(function.name())        
    def visitOperation(self, operation):
        self.strip(operation.name())
        for parameter in operation.parameters(): parameter.accept(self)


class Unduplicator(Visitor.AstVisitor, Visitor.TypeVisitor):
    """Visitor that removes duplicate declarations"""
    def __init__(self):
        self.__global = AST.MetaModule("",[])
	self.__scopes = [self.__global]
    def top_get_decl(self, name):
	for decl in self.__scopes[-1].declarations():
	    if hasattr(decl, 'name') and decl.name() == name:
		return decl
	return None
    def top_get_decl_list(self, name):
	decls = []
	for decl in self.__scopes[-1].declarations():
	    if hasattr(decl, 'name') and decl.name() == name:
		decls.append(decl)
	return decls
    def top_append(self, declaration):
	self.__scopes[-1].declarations().append(declaration)
    def push(self, scope):
	self.__scopes.append(scope)
    def pop(self):
	del self.__scopes[-1]
    def top(self):
	return self.__scopes[-1]
    def globalns(self):
	return self.__global
    def visitModule(self, module):
	# Check if should skip
	if not scopedNameOkay(module.name()):
	    #print "Skipping module",Util.ccolonName(module.name())
	    return
	metamodule = self.top_get_decl(module.name())
	if metamodule is None:
	    metamodule = AST.MetaModule(module.type(),module.name())
	    self.top_append(metamodule)
	metamodule.module_declarations().append(module)
	self.push(metamodule)
	decls = tuple(module.declarations())
	del module.declarations()[:]
        for decl in decls: decl.accept(self)
	self.pop()
    def visitMetaModule(self, module):
	# Check if should skip
	if not scopedNameOkay(module.name()):
	    #print "Skipping module",Util.ccolonName(module.name())
	    return
	metamodule = self.top_get_decl(module.name())
	if metamodule is None:
	    metamodule = AST.MetaModule(module.type(),module.name())
	    self.top_append(metamodule)
	metamodule.module_declarations().extend(module.module_declarations())
	self.push(metamodule)
	decls = tuple(module.declarations())
	del module.declarations()[:]
        for decl in decls: decl.accept(self)
	self.pop()
    def addDeclaration(self, decl):
	name = decl.name()
	if not scopedNameOkay(name): return
	# TODO: this is inefficient: O(n^2)
	for prev in self.top().declarations():
	    if prev.name() == name:
		return
	self.top().declarations().append(decl)
    def visitNamed(self, decl):
	name = decl.name()
	if self.top_get_decl(decl.name()): return
	self.addDeclaration(decl)
    visitDeclaration = addDeclaration
    visitForward = addDeclaration
    visitEnum = addDeclaration
    def visitFunction(self, func):
	if not scopedNameOkay(func.name()): return
	# Find any functions with same name and parameters
	#print "func.encoded:",func.encoded_name()
	if not isinstance(self.top(), AST.Class):
	    for decl in self.top().declarations():
		if not isinstance(decl, AST.Function): continue
		if func.name() == decl.name():
		    return
	ret = self.linkType(func.returnType())
	if ret is not func.returnType():
	    func.set_returnType(ret)
	for param in func.parameters():
	    self.visitParameter(param)
	self.top().declarations().append(func)
    visitOperation = visitFunction
    def visitVariable(self, var):
	if not scopedNameOkay(var.name()): return
	vt = self.linkType(var.vtype())
	if vt is not var.vtype():
	    var.set_vtype(vt)
	self.addDeclaration(var)
    def visitTypedef(self, tdef):
	if not scopedNameOkay(tdef.name()): return
	alias = self.linkType(tdef.alias())
	if alias is not tdef.alias():
	    tdef.set_alias(alias)
	self.addDeclaration(tdef)
    def visitClass(self, clas):
	if not scopedNameOkay(clas.name()): return
	if self.top_get_decl(clas.name()): return
	for parent in clas.parents():
	    parent.accept(self)
	self.push(clas)
	decls = tuple(clas.declarations())
	del clas.declarations()[:]
        for decl in decls: decl.accept(self)
	self.pop()
	self.addDeclaration(clas)
    def visitParameter(self, param):
	type = self.linkType(param.type())
	if type is not param.type():
	    param.set_type(type)
    def visitConst(self, const):
	if not scopedNameOkay(const.name()): return
	ct = self.linkType(const.ctype())
	if ct is not const.ctype():
	    const.set_ctype(ct)
	self.addDeclaration(const)

    ### TYPES
    def linkType(self, type):
	"Returns the same or new proxy type"
	self.__type = type
	if type is not None: type.accept(self)
	return self.__type
    def visitBaseType(self, type):
	if types.has_key(type.name()):
	    self.__type = types[type.name()]
    def visitForward(self, type):
	if types.has_key(type.name()):
	    self.__type = types[type.name()]
    def visitDeclared(self, type):
	if types.has_key(type.name()):
	    self.__type = types[type.name()]
	else:
	    print "Couldn't find declared type:",type.name()
    def visitTemplate(self, type):
	print "Unimplemented: visitTemplate:",type.__dict__
    def visitModifier(self, type):
	alias = self.linkType(type.alias())
	if alias is not type.alias():
	    type.set_alias(alias)
	self.__type = type
    def visitParametrized(self, type):
	params = tuple(type.parameters())
	del type.parameters()[:]
	for param in params:
	    type.parameters().append(self.linkType(param))
	self.__type = type
    def visitFunctionType(self, type):
	ret = self.linkType(type.returnType())
	if ret is not type.returnType():
	    type.set_returnType(ret)
	params = tuple(type.parameters())
	del type.parameters()[:]
	for param in params:
	    type.parameters().append(self.linkType(param))
	self.__type = type

def preprocess(types, declarations):
    global strip
    for scope in strip:
        stripper = Stripper(scope)
        for item in declarations:
            item.accept(stripper)

def scopedNameOkay(name):
    if not strip: return 1
    okay = 0
    for scope in strip:
	if Util.ccolonName(name)[:len(scope)] == scope:
	    okay = 1
	    break
    return okay
 
def stripTypes():
    if not strip: return
    for name in types.keys():
	try:
	    if not scopedNameOkay(name):
		type = types[name]
		if not isinstance(type, Type.Forward):
		    types[name] = Type.Forward(type.language(), name)
	except:
	    print "ERROR Processing:",name,types[name]
	    raise


def resolve(args, types_, declarations):
    global types
    types = types_

    __parseArgs(args)
    #preprocess(types, declarations)
    #
    # for now, avoid all declarations not from a main file, i.e.
    # implicite (by means of included files) declarations

    stripTypes()

    unduper = Unduplicator()

    for decl in declarations:
	decl.accept(unduper)

    declarations[:] = unduper.globalns().declarations()

    # Remove types not in strip
    declarations = filter(lambda d: d.mainfile(), declarations)
    
